"""Functions for working with slicer files"""

import json
import logging
import re
from typing import Tuple

import numpy as np
import SimpleITK as sitk

from aind_mri_utils.coordinate_systems import convert_coordinate_system

from ..sitk_volume import (
    find_points_equal_to,
    transform_sitk_indices_to_physical_points,
)
from ..utils import find_indices_equal_to

logger = logging.getLogger(__name__)


def extract_control_points(json_data: dict) -> Tuple[np.ndarray, list]:
    """
    Extract points and names from slicer json dict

    Parameters
    ----------
    json_data : dict
        Contents of json file

    Returns
    -------
    pts : numpy.ndarray (N x 3)
        point positions
    labels : list
        labels of controlPoints
    coord_str : str
        String specifying coordinate system of pts, e.g. 'LPS'
    """
    pts = json_data["markups"][0]["controlPoints"]
    coord_str = json_data["markups"][0]["coordinateSystem"]
    labels = []
    pos = []
    for ii, pt in enumerate(pts):
        labels.append(pt["label"])
        pos.append(pt["position"])
    return np.array(pos), labels, coord_str


def find_seg_nrrd_header_segment_info(header):
    """
    parse keys of slicer created dict to find segment names and values


    Parameters
    ----------
    header : dict-like

    Returns
    -------
    segment_info: dict
        pairs of segment name : segment value
    """
    matches = filter(
        None,
        map(lambda s: re.match("^([^_]+)_LabelValue$", s), header.keys()),
    )
    segment_info = dict()
    for m in matches:
        segment_name = header["{}_Name".format(m[1])]
        segment_info[segment_name] = int(header[m[0]])
    return segment_info


def get_segmented_labels(label_vol):
    """
    Extract metadata from the implant volume and return the segmentation label
    dictionary.

    Parameters
    ----------
    label_vol : SimpleITK.Image
        The label volume from which to extract segmentation information.

    Returns
    -------
    dict
        A dictionary mapping segmentation labels to their corresponding values
        in the volume.
    """
    metadata_dict = {
        k: label_vol.GetMetaData(k) for k in label_vol.GetMetaDataKeys()
    }
    label_dict = find_seg_nrrd_header_segment_info(metadata_dict)
    return label_dict


def load_segmentation_points(label_vol, order=None, image=None):
    """
    Load segmentation points from a 3D Slicer generated .seg.nrrd file

    Note that, because this is effectively an image loader, it is excluded from
    coverage tests

    Parameters
    ----------
    label_vol : SimpleITK.Image or str
        SimpleITK.Image or filename to open. Must be .seg.nrrd
    order : list of strings, optional
        list of segment names to load.
        Labels will be in order specified by order.
        If None, labels will be loaded in the order they are found in the file.
        Default is None.
    image : SimpleITK.Image, optional
        Image to use for extracting the weights
        (image intensity) for each labeled point

    Returns
    -------
    positions : np.array(N,3)
        xyz positions of the labeled points
    labels : np.array(N,)
        labels of the labeled points
    weights : np.array(N,)  (optional)
        weights of the labeled points. Only returned if image is not None.

    """
    # Read the volume if a string is passed
    if isinstance(label_vol, str):
        if not label_vol.endswith(".seg.nrrd"):
            raise ValueError("label_vol must be a .seg.nrrd file")
        label_vol = sitk.ReadImage(label_vol)

    # Get the labels from the header
    label_dict = get_segmented_labels(label_vol)

    if order is None:
        order = list(label_dict.keys())

    labels = []
    positions = []
    if image is None:
        for label_ndx, label in enumerate(order):
            these_positions = find_points_equal_to(
                label_vol, label_dict[label]
            )
            label_ndxs = np.full(these_positions.shape[0], label_ndx)
            positions.append(these_positions)
            labels.append(label_ndxs)
        return np.concatenate(positions), np.concatenate(labels)
    else:
        weights = []
        label_arr = sitk.GetArrayViewFromImage(label_vol)
        img_arr = sitk.GetArrayViewFromImage(image)
        for label_ndx, label in enumerate(order):
            val = label_dict[label]
            ndxs = find_indices_equal_to(label_arr, val)
            sitk_ndxs = ndxs[:, ::-1]
            these_positions = transform_sitk_indices_to_physical_points(
                label_vol, sitk_ndxs
            )
            label_ndxs = np.full(these_positions.shape[0], label_ndx)
            np_ndxs = tuple(ndxs.T)
            these_weights = img_arr[np_ndxs]
            positions.append(these_positions)
            weights.append(these_weights)
            labels.append(label_ndxs)
        return (
            np.concatenate(positions),
            np.concatenate(labels),
            np.concatenate(weights),
        )


def markup_json_to_numpy(filename):
    """
    Extract control points from a 3D Slicer generated markup JSON file

    Parameters
    ----------
    filename : string
        filename to open. Must be .json
        .mrk.json is ok
    Returns
    -------
    pts, names - numpy.ndarray (N x 3) of point positions and list of
    controlPoint names

    """
    with open(filename) as f:
        data = json.load(f)
    return extract_control_points(data)


def markup_json_to_dict(filename):
    """
    Extract control points from a 3D Slicer generated markup JSON file

    Parameters
    ----------
    filename : string
        filename to open. Must be .json
        .mrk.json is ok

    Returns
    -------
    Dictionary
        dictionary with keys = point names and values = np.array of points.


    """
    pos, names, coord_string = markup_json_to_numpy(filename)
    return dict(zip(names, pos)), coord_string


def create_slicer_fcsv(filename, pts_dict, direction="LPS"):
    """
    Save fCSV file that is slicer readable.
    """
    # Create output file
    with open(filename, "w+") as f:
        f.write("# Markups fiducial file version = 4.11\n")
        f.write("# CoordinateSystem = {}\n".format(direction))
        f.write(
            "# columns = id,x,y,z,ow,ox,oy,oz,vis,sel,lock,label,desc,associatedNodeID\n"  # noqa: E501
        )

        for pt_no, key in enumerate(pts_dict.keys()):
            x, y, z = pts_dict[key]
            f.write(
                "{:d},{:f},{:f},{:f},0,0,0,1,1,1,0,{!s},,vtkMRMLScalarVolumeNode1\n".format(  # noqa: E501"
                    pt_no + 1, x, y, z, key
                )
            )


def _parse_slicer_fcsv_header(file):
    "Parse the header of a slicer fcsv file, returning the last line read"
    line = file.readline()
    source_coordinate_system = None
    column_headers = []
    # Parse the header
    while line and line.startswith("#"):
        key, _, value = line.partition("=")
        value = value.strip()
        if "CoordinateSystem" in key:
            source_coordinate_system = value
        elif "columns" in key:
            column_headers = [col.strip() for col in value.split(",")]
        line = file.readline()
    if not column_headers:
        raise ValueError("No column headers found in file")
    if source_coordinate_system is None:
        raise ValueError("No CoordinateSystem found in file")
    col_ndxs = {col: i for i, col in enumerate(column_headers)}
    if "label" not in col_ndxs:
        raise ValueError("No 'label' column found in file")

    # Get the indices of the needed columns
    label_ndx = col_ndxs["label"]
    coord_ndxs = [col_ndxs[axis] for axis in ["x", "y", "z"]]
    last_line = line

    return last_line, source_coordinate_system, label_ndx, coord_ndxs


def read_slicer_fcsv(filename, direction="LPS"):
    """
    Read fscv into dictionary.
    While reading, points will be converted to the specified direction.

    Parameters
    ----------
    filename : string
        filename to open. Must be .fcsv
    direction : string (optional)
        direction of the coordinate system of the points in the file.
        Must be one of 'LPS','RAS','LAS','LAI','RAI','RPI','LPI','LAI'
        Default is 'LPS'

    Returns
    -------
    Dictionary
        dictionary with keys = point names and values = np.array of points.
    """
    if not filename.endswith(".fcsv"):
        raise ValueError("File must be a .fcsv file")
    valid_directions = {"LPS", "RAS", "LAS", "LAI", "RAI", "RPI", "LPI", "LAI"}
    if direction not in valid_directions:
        raise ValueError(f"Direction must be one of {valid_directions}")
    logger.debug(f"Reading {filename} with direction {direction}")

    point_dictionary = {}

    with open(filename, "r") as file:
        line, source_coordinate_system, label_ndx, coord_ndxs = (
            _parse_slicer_fcsv_header(file)
        )
        need_conversion = source_coordinate_system != direction

        while line:
            point_parts = line.strip().split(",")
            point_key = point_parts[label_ndx]
            point_values = np.array(
                [float(point_parts[i]) for i in coord_ndxs]
            )
            if need_conversion:
                point_values = convert_coordinate_system(
                    point_values, source_coordinate_system, direction
                )
            point_dictionary[point_key] = point_values
            line = file.readline()

    return point_dictionary



"""
Functions for correcting for chemical shift in MRI images
"""

import numpy as np


def compute_chemical_shift(
    image, ppm=(3.7 + 4.1) / 2, mag_freq=599, pixel_bandwidth=500
):
    """Calculate the chemical shift for an MRI image.

    The chemical shift is calculated based on the parts per million (ppm),
    magnetic frequency of the scanner (mag_freq), and per-pixel bandwidth
    of the image.

    Parameters
    ----------
    image : SimpleITK.Image
        The input image, which is an instance of a SimpleITK.Image object.
    ppm : float, optional
        The parts per million value for the chemical shift calculation.
        Default is the average of 3.7 and 4.1.
    mag_freq : float, optional
        The magnetic frequency in MHz. Default is 599.
    pixel_bandwidth : float, optional
        The pixel bandwidth in Hz. Default is 500.

    Returns
    -------
    float
        The computed chemical shift value.
    """
    shift = (ppm * mag_freq) / pixel_bandwidth * image.GetSpacing()[1]
    return shift


def chemical_shift_transform(shift, readout="HF"):
    """Create chemical shift transformation matrix.

    Creates a transformation matrix that accounts for the chemical
    shift in MRI images. The direction of the readout (either head-foot (HF) or
    left-right (LR)) determines the configuration of the transformation matrix.

    Parameters
    ----------
    shift : float
        The chemical shift value to be applied.
    readout : str, optional
        The direction of the readout. It can be either 'HF' for head-foot
        direction or 'LR' for left-right direction.  Default is 'HF'.

    Returns
    -------
    R : np.ndarray
        A 3x3 rotation matrix.
    translation : np.ndarray
        A 3-element translation

    Raises
    ------
    ValueError
        If the readout direction is not recognized.
    """
    if readout == "HF":
        translation = np.array([0, shift, 0])
    elif readout == "LR":
        translation = np.array([shift, 0, 0])
    else:
        raise ValueError("Readout direction not recognized")
    R = np.eye(3)
    return R, translation



"""Tests functions in `file_io.obj_files`."""

import inspect
import unittest
from unittest.mock import MagicMock, patch

import numpy as np
import pywavefront

from aind_mri_utils.file_io import obj_files as of


class ObjFilesTest(unittest.TestCase):
    """Tests functions in `file_io.obj_files`."""

    scene_mock = MagicMock()
    scene_mock.vertices = [
        (-6.859, 0.7264, 2.84),
        (-6.859, 1.5264, 2.84),
        (-8.059, 1.5264, 2.84),
        (-8.049885, 1.5264, 2.944189),
        (-8.022816, 1.5264, 3.045212),
    ]
    scene_mock.mesh_list = [MagicMock(), MagicMock()]
    scene_mock.mesh_list[0].faces = [
        [0, 2, 3],
        [0, 3, 4],
        [2, 3, 4],
        [1, 2, 4],
        [1, 3, 4],
    ]
    scene_mock.mesh_list[1].faces = scene_mock.mesh_list[0].faces

    expected_vertices = np.array(scene_mock.vertices)
    expected_faces_per_mesh = np.array(scene_mock.mesh_list[0].faces)

    def test_load_obj_wavefront(self) -> None:
        """Tests that the `load_obj_wavefront` function works as intended."""
        # inspect API for Wavefront to see that it's roughly what we need
        s = inspect.signature(pywavefront.Wavefront)
        self.assertTrue("strict" in s.parameters)
        self.assertTrue("create_materials" in s.parameters)
        self.assertTrue("collect_faces" in s.parameters)
        # Call the function to achieve 100% coverage (why though?)
        #
        # To be clear this is just for the coverage, and is not a meaningful
        # test
        with patch(
            "aind_mri_utils.file_io.obj_files.pywavefront.Wavefront"
        ) as mock:
            mock.return_value = True
            self.assertTrue(of.load_obj_wavefront("foobar"))

    def test_get_vertices_and_faces(self) -> None:
        """Tests `get_vertices_and_faces`"""
        vertices, faces = of.get_vertices_and_faces(self.scene_mock)
        self.assertTrue(np.array_equal(vertices, self.expected_vertices))
        self.assertEqual(len(faces), 2)
        self.assertTrue(np.array_equal(faces[1], self.expected_faces_per_mesh))
        with patch(
            "aind_mri_utils.file_io.obj_files.pywavefront.Wavefront"
        ) as mock:
            mock.return_value = self.scene_mock
            vertices, faces = of.get_vertices_and_faces("foobar")
            self.assertTrue(np.array_equal(vertices, self.expected_vertices))
            self.assertEqual(len(faces), 2)
            self.assertTrue(
                np.array_equal(faces[1], self.expected_faces_per_mesh)
            )


if __name__ == "__main__":
    unittest.main()



"""
IO functions for SITK
"""

import os

import SimpleITK as sitk


def read_image(filename):
    """
    Reads generic image files/folders in SITK using
    Currently explicitly supported: .dcm, .nii, .tiff
    Folders/stacks will be read for .dcm and .tiff files
    Other formats work only if supported they work with sitk.ReadImage()

    Parameters
    ----------
    filename : String
        filename or folder of files.

    Returns
    -------
    SITK image
        SITK image from loaded dicom files.

    """

    if os.path.isdir(filename):
        # Look for .tifs
        tiff_list = [
            x
            for x in os.listdir(
                filename,
            )
            if (".tif" in x)
        ]
        # Read tifs if they exist
        if len(tiff_list) > 0:
            return read_tiff_stack(filename)
        else:
            # in the absence of tifs, assume folders are dcm
            return read_dicom(filename)

    else:
        if (".nii" in filename) or (".nifti" in filename):
            return read_nii(filename)
        elif (".dcm" in filename) or (
            os.path.splitext(filename)[0] == filename
        ):
            return read_dicom(filename)
        else:
            # If none of the conditions above are reached, try to
            # use the default reader. This will throw an error if there
            # are any problems
            return sitk.ReadImage(filename)


def read_dicom(filename):
    """Reader to import Dicom file and convert to sitk image

    See
    https://simpleitk.readthedocs.io/en/master/link_DicomSeriesReader_docs.html#lbl-dicom-series-reader

    Parameters
    ----------
    filename : String
        folder of .dcm image. If an individual image file name is passes, will
        read all .dcm files in that folder

    Returns
    -------
    SITK image
        SITK image from loaded dicom files.

    """

    if os.path.isdir(filename):
        dirname = filename
    else:
        dirname = os.path.dirname(filename)

    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(
        dirname, useSeriesDetails=True, loadSequences=True
    )
    reader.SetFileNames(dicom_names)
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()
    return reader.Execute()


def read_dcm(filename):
    """
    Reader to import Dicom file and convert to sitk image.
    This function is a wrapper on read_dicom to handle multiple naming
    conventions

    Parameters
    ----------
    filename : String
        folder of .dcm image. If an individual image file name is passes, will
        read all .dcm files in that folder

    Returns
    -------
    SITK image
        SITK image from loaded dicom files.

    """
    return read_dicom(filename)


def read_nii(filename):
    """
    Reader to import nifti file and convert to sitk image
    This function is just a wrapper to match convention.

    Parameters
    ----------
    filename : String
        filename of .nii file.

    Returns
    -------
    SITK image
        SITK image from loaded dicom files..

    """
    return sitk.ReadImage(filename)


def read_nifti(filename):
    """
    Reader to import nifti file and convert to sitk image
    This function is a wrapper on read_nii to handle multiple naming
    conventions, which is in turn just an sitk wrapper.

    Parameters
    ----------
    filename : String
        filename of .nii file.

    Returns
    -------
    SITK image
        SITK image from loaded dicom files..

    """
    return read_nii(filename)


def read_tiff_stack(folder):
    """
    Code to read a tiff stack
    THIS CODE IS INCOMPLETE: needs metadata handling (resolution, etc.) and
    some thought about how to deal with large images.

    Parameters
    ----------
    folder : String folder with numerically ordered tiff images
        Folder containing ONLY .tif images

    Returns
    -------
    SITK image
        Tiff images stacked.

    """
    reader = sitk.ImageSeriesReader()
    lst = [
        x
        for x in os.listdir(
            folder,
        )
        if (".tif" in x)
    ]
    lst = [os.path.join(folder, x) for x in lst]
    reader.SetFileNames(lst)
    return reader.Execute()


def write_nii(image, filename):
    """

    Write an sitk image to .nii file


    Parameters
    ----------
    image : SITK image
        Image to save.

    filename : filename
        Filename to save

    """
    f_name, f_ext = os.path.splitext(filename)
    if len(f_ext) == 0:
        f_ext = ".nii"
    elif f_ext != ".nii":
        raise Exception("Filename should be a .nii file")
    sitk.WriteImage(image, f_name + f_ext)


def write_dicom(image, foldername):
    """
    Save SITK image as dicom stack.
    Heavily borrowed from
    https://simpleitk.readthedocs.io/en/master/link_DicomSeriesReadModifyWrite_docs.html

    Parameters
    ----------
    image : SITK image
        Image to save.

    filename : filename
        Filename to save

    """
    # Need to figure out options to copy metadata
    raise NotImplementedError(
        "Dicom file writer still needs to be implemented"
    )



"""Plotting functions"""

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import tri as mpt
from matplotlib.colors import ListedColormap


def plot_tri_mesh(ax, vertices, faces, *plot_args, **plot_kwargs):
    """
    Adds a mesh to a 3d plot

    Parameters
    ==========
    ax - 3d axis to plot on
    vertices - N x 3 ndarray of coordinates for N vertices
    faces - N x 3 ndarray of vertex indices defining triangular faces
    *plot_args - varargs passed to plot call
    **plot_kwargs - keyword args passed to plot call

    Returns
    =======
    handles, tri - handles to plot polygons, and triangulation
    """
    tri = mpt.Triangulation(vertices[:, 0], vertices[:, 1], triangles=faces)
    handles = ax.plot_trisurf(tri, vertices[:, 2], *plot_args, **plot_kwargs)
    return handles, tri


# Function from @Mateen Ulhaq and @karlo
def _set_axes_radius(ax, origin, radius):
    """
    Set all three axes to have the same distance around the origin

    Parameters
    ==========
    ax - matplotlib axis handle
    origin - np.ndarray (3) specifying origin point
    radius - scalar radius around origin used to set axis limits
    """

    x, y, z = origin
    ax.set_xlim3d([x - radius, x + radius])
    ax.set_ylim3d([y - radius, y + radius])
    ax.set_zlim3d([z - radius, z + radius])


# Function from @Mateen Ulhaq and @karlo
def set_axes_equal(ax: plt.Axes):
    """Set 3D plot axes to equal scale.

    Make axes of 3D plot have equal scale so that spheres appear as
    spheres and cubes as cubes.  Required since `ax.axis('equal')`
    and `ax.set_aspect('equal')` don't work on 3D.
    """
    limits = np.array(
        [
            ax.get_xlim3d(),
            ax.get_ylim3d(),
            ax.get_zlim3d(),
        ]
    )
    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    _set_axes_radius(ax, origin, radius)


def make_3d_ax_look_normal(ax: plt.Axes):
    """
    Changes the aspect ratio of a 3d plot so that dimensions are approximately
    the same size

    Parameters
    ==========
    ax - matplotlib 3d axis
    """
    ax.set_box_aspect([1, 1, 1])
    set_axes_equal(ax)


def get_prop_cycle():
    """
    Returns the colors in the current prop cycle
    """
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    return prop_cycle.by_key()["color"]


def plot_point_cloud_3d(a, pts, *args, **kwargs):
    """Scatter plots a Nx3 array as N 3D points"""
    return a.scatter(pts[:, 0], pts[:, 1], pts[:, 2], *args, **kwargs)


def create_single_colormap(
    color_name,
    N=256,
    saturation=0,
    start_color="white",
    is_transparent=True,
    is_reverse=False,
):
    """ "
    Creates a colormap with a single color

    Parameters
    ==========
    color_name - string name of color
    N - number of colors in colormap
    saturation - number of colors to add to the colormap
    start_color - color to start the colormap with
    is_transparent - whether to make the start color transparent
    is_reverse - whether to reverse the colormap
    Returns
    =======
    cmap - matplotlib colormap

    """
    cmap = ListedColormap([start_color, color_name])
    start_color = np.array(cmap(0))
    if is_transparent:
        start_color[-1] = 0
    if not is_reverse:
        cmap = ListedColormap(
            np.vstack(
                (
                    np.linspace(start_color, cmap(1), N),
                    np.tile(cmap(1), (int(saturation * N), 1)),
                )
            )
        )
    else:
        cmap = ListedColormap(
            np.vstack(
                (
                    np.tile(cmap(1), (int(saturation * N), 1)),
                    np.linspace(cmap(1), start_color, N),
                )
            )
        )
    return cmap


def plot_vector(a, pt, *args, **kwargs):
    """Plots a 3D point as a vector from the origin"""
    plt_pts = np.vstack([np.array([0, 0, 0]), pt])
    return a.plot(plt_pts[:, 0], plt_pts[:, 1], plt_pts[:, 2], *args, **kwargs)


def rgb_to_int(r, g, b):
    """Converts an RGB color to an integer.

    Parameters
    ----------
    r : int
        Red component of the color, must be in the range [0, 255].
    g : int
        Green component of the color, must be in the range [0, 255].
    b : int
        Blue component of the color, must be in the range [0, 255].

    Returns
    -------
    int
        The integer representation of the RGB color.

    Raises
    ------
    ValueError
        If any of the RGB values are not in the range [0, 255].

    Examples
    --------
    >>> rgb_2_int(255, 0, 0)
    16711680
    >>> rgb_2_int(0, 255, 0)
    65280
    >>> rgb_2_int(0, 0, 255)
    255

    """
    out = 0
    for i, v in enumerate((b, g, r)):
        if v < 0 or v > 255:
            raise ValueError(
                f"RGB values must be in the range [0, 255], got {v}"
            )
        out |= v << (8 * i)
    return out


def rgb_to_hex_string(r, g, b):
    """Converts an RGB color to a hex string.

    Parameters
    ----------
    r : int
        Red component of the color (0-255).
    g : int
        Green component of the color (0-255).
    b : int
        Blue component of the color (0-255).

    Returns
    -------
    str
        Hexadecimal string representation of the color.

    """
    color_int = rgb_to_int(r, g, b)
    return "0x{0:06X}".format(color_int)


def hex_string_to_int(hx):
    """Converts a hex color string to an integer.

    Parameters
    ----------
    hx : str
        Hexadecimal string representation of the color.

    Returns
    -------
    int
        The integer representation of the hex color.

    Examples
    --------
    >>> hex_string_2_int("#FF0000")
    16711680
    >>> hex_string_2_int("0x00FF00")
    65280
    >>> hex_string_2_int("0000FF")
    255

    """
    return int(hx.lstrip("#"), 16)



"""Utility functions"""

import math

import numpy as np
from numpy import linalg


def skew_symmetric_cross_product_matrix(v):
    """Find the cross product matrix for a vector v"""
    return np.cross(v, np.identity(v.shape[0]) * -1)


def norm_vec(vec):
    """Normalize input vector"""
    n = np.linalg.norm(vec)
    if n == 0:
        raise ValueError("Input has norm of zero")
    return vec / n


def vector_rejection(v, n):
    """Find the component of v orthogonal to n"""
    ndim = n.size
    nn = norm_vec(n)
    vn = (v.reshape(-1, ndim) @ nn[:, np.newaxis]) * nn[np.newaxis, :]
    return v - vn


def mask_arr_by_annotations(arr, anno_arr, seg_vals, default_val=0):
    """Sets entries of arr to default_val if anno_arr not in target set

    This function will return a copy of `arr` where the output is either
    the same as `arr` if the corresponding element of `anno_arr` is one of
    `seg_vals`, or `default_val` if not.

    Parameters
    ----------
    arr : numpy.ndarray
        Array that will be masked
    anno_arr : numpy.ndarray
        Array same size as `arr` that assigns each element to a segment
    seg_vals : set like
        Set of values that anno_arr will be compared to
    default_val : number
        value of output array if anno_arr is not in seg_vals, default = 0.

    Returns
    -------
    masked_vol : numpy.ndarray
        Copy of `arr` masked by whether `anno_arr` is one of `seg_vals`
    """

    masked_arr = np.zeros_like(arr)
    masked_arr.fill(default_val)
    mask = np.isin(anno_arr, seg_vals)
    masked_arr[mask] = arr[mask]
    return masked_arr


def find_indices_equal_to(arr, v):
    """Find array indices equal to v"""
    return np.column_stack(np.nonzero(arr == v))


def get_first_pca_axis(pts):
    """Find first PC of points"""
    centered = pts - np.mean(pts, axis=0)[np.newaxis, :]
    _, _, vh = linalg.svd(centered, full_matrices=False)
    return vh[0, :]


def signed_angle_rh(a, b, n):
    """find right-handed angle between two vectors
    Find the right-handled angle between a and b in the plane normal to n,
    by rotating a to b
    """
    # Function by Adrian Leonhard: https://stackoverflow.com/a/33920320
    # Let alpha be the direct angle between the vectors (0° to 180°) and beta
    # the angle we are looking for (0° to 360°) with beta == alpha or
    # beta == 360° - alpha
    #
    # Va . Vb == |Va| * |Vb| * cos(alpha) (by definition)
    #         == |Va| * |Vb| * cos(beta)
    # As cos(alpha) == cos(-alpha) == cos(360° - alpha)
    #
    #
    # Va x Vb == |Va| * |Vb| * sin(alpha) * n1
    # (by definition; n1 is a unit vector perpendicular to Va and Vb with
    # orientation matching the right-hand rule)
    #
    # Therefore (again assuming Vn is normalized):
    # n1 . Vn == 1 when beta < 180
    # n1 . Vn == -1 when beta > 180
    #
    # ==>  (Va x Vb) . Vn == |Va| * |Vb| * sin(beta)
    # Finally,
    # tan(beta) = sin(beta) / cos(beta) == ((Va x Vb) . Vn) / (Va . Vb)

    vn = norm_vec(n)
    return math.atan2(np.dot(np.cross(a, b), vn), np.dot(a, b))


def signed_angle_lh(a, b, n):
    """find left-handed angle between two vectors

    See `signed_angle_rh`
    """
    return signed_angle_rh(b, a, n)


def unsigned_angle(a, b):
    """
    Calculate the unsigned angle between two vectors.

    Parameters
    ----------
    a : array-like
        First vector.
    b : array-like
        Second vector.

    Returns
    -------
    angle : float
        The unsigned angle between the two vectors in radians.
    """
    an = norm_vec(a)
    bn = norm_vec(b)
    return np.arccos(np.clip(np.dot(an, bn), -1.0, 1.0))
